package nl.uvt.locator;

import java.net.*;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Properties;

import nu.xom.Document;
import nu.xom.Nodes;

public class LinkDbConnection {
	List<LinkInfo> query( QueryData pQ ) throws LocatorException{
		List<LinkInfo> linkInfo = null;
		filterForgetAbouts( pQ );
		if ( pQ.issn.length() > 0 ){
			linkInfo = queryLinkDbIssn( pQ );   
		}
		else if ( pQ.title.length() > 0 ){
			linkInfo = queryLinkDbTitle( pQ );   
		}
		return linkInfo;
	}
	
	private static void filterForgetAbouts( QueryData pQ ) {
		if ( Locator.forgetAbout == null ){
			Locator.forgetAbout = new HashSet<String>();
			String words = Locator.locatorConfig.getProperty("forgetAbout","");
			if ( !words.equals("") ){
				String [] doiS = words.split(",");
				for ( int i=0; i < doiS.length; ++i ){
					Locator.forgetAbout.add(doiS[i].trim());
				}
			}
			Locator.logger.debug("filled forgetAbout list: " + Locator.forgetAbout.toString() );
		}
		Iterator<String> it = Locator.forgetAbout.iterator();
		while ( it.hasNext() ){
			String what = it.next();
			Locator.logger.debug("try to Forget all about '" + what + "'" );
			if ( what.equals("issn") )
				pQ.issn = "";
			else if ( what.equals("isbn") )
				pQ.issn = "";
			else if ( what.equals("year") )
				pQ.year = -1;
			else if ( what.equals("spage") )
				pQ.spage = "";
			else if ( what.equals("aufirst") )
				pQ.aufirst = "";
			else if ( what.equals("aufirst") )
				pQ.aufirst = "";
			else if ( what.equals("aufirst") )
				pQ.aufirst = "";
			else if ( what.equals("aulast") )
				pQ.aulast = "";
			else if ( what.equals("atitle") )
				pQ.atitle = "";
			else if ( what.equals("title") )
				pQ.title = "";
			else if ( what.equals("auinit1") )
				pQ.auinit1 = "";
			else if ( what.equals("genre") )
				pQ.genre = "";
			else if ( what.equals("issue") ){
				pQ.issue = "";
				pQ.thIssue = -1;
			}
			else if ( what.equals("location") )
				pQ.location = "";
			else if ( what.equals("pages") )
				pQ.pages = "";
			else if ( what.equals("volume") ){
				pQ.volume = "";
				pQ.thVolume = -1;
			}
			else
				Locator.logger.warn("Unable to forget '" + what + "' unknown or unimplemented value!");
		}
	}


	private List<LinkInfo> queryLinkDbIssn( QueryData pQ ) throws LocatorException {
		List<LinkInfo> linkResult = null;
		nu.xom.Document result;
		String linkDbUrl = Locator.locatorConfig.getProperty("linkDbUri");
		String query = "";
		try {
			// try linkDb with issn
			Properties urlProps = new Properties( Locator.sruDefaultProps );
			query = URLEncoder.encode( pQ.issn, "ISO-8859-1" );
			urlProps.setProperty("query", "bath.issn="+query );
			linkDbUrl = HttpConnection.createUrlString( linkDbUrl, urlProps );
			Locator.logger.debug("Try URL: " + linkDbUrl );
			result = HttpConnection.httpGetDocument( linkDbUrl, Locator.httpConfig );
			if ( result != null ){
			Locator.logger.debug( "ISSN search returns:");
//			Locator.logger.debug( result.toXML() );
				int hits = Locator.extractHits( result );
				Locator.logger.debug("linkdb numOfRecords=" + hits );
				if ( hits > 0 ){
					linkResult = getLinkDbResults( result );
				}
			}
			else {
				Locator.logger.warn("No response from:" + linkDbUrl );
			}
		}
		catch (Exception e)  {
			throw new LocatorException( "LinkDb URL: '" + linkDbUrl + "' failed: " + e.getMessage());
		}
		return linkResult;
	}

	private  List<LinkInfo> queryLinkDbTitle( QueryData pQ ) throws LocatorException {
		List<LinkInfo> linkResult = null;
		nu.xom.Document result;
		String linkDbUrl = Locator.locatorConfig.getProperty("linkDbUri");
		try {
			// try linkDb with Title
			Properties prop = new Properties( Locator.sruDefaultProps );
			String titleParts[] = pQ.title.split( ":", 2 );
			String mainTitle = pQ.title;
			if ( titleParts.length > 1 ){
				for( int i=0; i < titleParts.length; ++i )
					Locator.logger.debug( "part["+i+"]= '"+titleParts[i] +"'");
				mainTitle = titleParts[0];
			}
			prop.setProperty("query", "dc.title=\"^" + mainTitle + "^\"" );
			linkDbUrl = HttpConnection.createUrlString(linkDbUrl , prop);
			Locator.logger.debug("Try URL: " + linkDbUrl );
			result = HttpConnection.httpGetDocument( linkDbUrl, Locator.httpConfig);
//			logger.debug( "linkDb Title search returns:");
//			logger.debug( result.toXML() );
			int hits = Locator.extractHits( result );
			Locator.logger.debug("linkdb numOfRecords=" + hits );
			if ( hits > 0 ){
				linkResult = getLinkDbResults( result );
			}
		}
		catch (Exception e) 
		{
			throw new LocatorException( "LinkDb URL:'" + linkDbUrl + "' failed: "+ e.getMessage());
		}
		return linkResult;
	}

	private List<LinkInfo> getLinkDbResults( Document doc ){
		List<LinkInfo> linkResult = null;
		Nodes records = doc.query("//srw:recordData/record|//srw:recordData/rec", Locator.sruContext);
		if ( records.size() > 0 ){
			linkResult = new LinkedList<LinkInfo>();
			for ( int i=0; i < records.size(); ++i ){
				LinkDbInfo newInfo = new LinkDbInfo();
				newInfo.fill( records.get(i) );
				Locator.logger.debug( newInfo.toString());
				linkResult.add( newInfo );
			}
		}
		else{
			Locator.logger.warn("no record fields found!");
			Locator.logger.warn(doc.toXML());
		}
		return linkResult;
	}


}
